#!/bin/bash
#
# all main functions to generate a PKI certificate chain
# 

#
# Set the CA variables
#
pki_func_init() {
  if [[ -n $1 ]] || [[ -n $2 ]] || [[ -n $3 ]]; then
    FQ_CA_CERT=$1
    FQ_CA_KEYS=$2
    CNF_PATH=$3
    APP_INIT=1
  else
    APP_INIT=0
  fi
}

#
# print text wrapped in a block
#
echo_block() {
  echo
  echo "***** ***** ***** *****"
  echo $1
  echo "***** ***** ***** *****"
}

# 
# Grab the latest serial # from the file, auto-increment
# 
get_serial() {
  SERIAL=`head "docs/SERIAL_LC"`
  if [[ -z $SERIAL ]]; then
    SERIAL=11111
    echo_block "WARN: no file 'SERIAL' found, using default 11111 as the serial # for CA"
  fi
}

# ***** ***** ***** ***** *****
# 
# CERTIFICATE AUTHORITY (CA)
# 
# ***** ***** ***** ***** *****
# This function will generate a CA Intermediate
# IN: UNIQ_ID_CA, SERIAL
#
gen_ca() {
  UNIQ_ID_CA=$1
  SERIAL=$2

  echo_block "Create CA (${UNIQ_ID_CA})"

  # encrypt the key
  #openssl genrsa -aes256 -out ca.keys.pem 4096
  #openssl genrsa -aes256 -password "pass:password" -out ca.keys.pem 4096 

  # key un-protected
  openssl genrsa -out "ca_${UNIQ_ID_CA}.keys.pem" 4096
  # 
  # Create Certificate   (valid for 10 years, after the entire chain of trust expires)
  openssl req -config $CNF_PATH/ca.cnf -new -x509 -sha256 -days 3650 -extensions v3_ca \
              -subj "/C=OO/O=ACME/CN=root.${UNIQ_ID_CA}" -set_serial ${SERIAL} \
              -key ca_${UNIQ_ID_CA}.keys.pem -out ca_${UNIQ_ID_CA}.crt.pem

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in ca_${UNIQ_ID_CA}.crt.pem > ca_${UNIQ_ID_CA}_cert.info.txt
}

#
# Create CA Intermediate PKI
# 
# 
#
# INPUT:  SERIAL #, LOOP NUM
# 
ca-i_gen_pki() {
  # organization
  CDD=`pwd`
  ORG_URL=$1
  SERIAL=$2
  LOOP_NUM=$3

  UNIQ_DIR_CA="ca_i_${SERIAL}.${ORG_URL}"
  mkdir -p "cfg"
  mkdir -p "distribution/${UNIQ_DIR_CA}"
  cd "distribution/${UNIQ_DIR_CA}"

  # geneate certificates, organize the files
  ca-i_gen_pki_certs $ORG_URL $SERIAL $LOOP_NUM
  ca-i_organize
  ca-i_cp_docs

  # return to last path
  cd $CDD
}

#
# Generate a PKI chain
#   - the certificate chain is unique based on the serial #
#   - generate a new CA I
#   - generate two server certificates
#   - generate two client certificates
#
# INPUT:  BASE SERIAL #, LOOP NUM
# 
# Requires: FQ_CA_CERT, FQ_CA_KEYS
# 
ca-i_gen_pki_certs() {
  ORG_URL=$1
  SERIAL_O=$2
  NUM_CERTS=$(($3-1))

  # Create CA Intermediate
  UNIQ_ID_CAI="${SERIAL_O}.${ORG_URL}" 
  ca-i_gen_cert $UNIQ_ID_CAI $SERIAL_O

  # Server Certificates
  for NUM in $(seq 0 $NUM_CERTS)
  do
    gen_server $ORG_URL $UNIQ_ID_CAI $((SERIAL_O+NUM))
  done

  # Client Certificates
  for NUM in $(seq 0 $NUM_CERTS)
  do
    gen_client $ORG_URL $UNIQ_ID_CAI $((SERIAL_O+NUM))
  done
}

# This function will generate a CA Intermediate
# 
# Requires:  CNF file, CA cert, CA key
# 
# IN: UNIQ_ID_CA, SERIAL
#
ca-i_gen_cert() {
  UNIQ_ID_CAI=$1
  SERIAL=$2

  echo_block "Create CA Intermediate  (${UNIQ_ID_CAI})"

  openssl genrsa -out "ca_i_${UNIQ_ID_CAI}.keys.pem" 4096

  # Create Cert Signing Request (CSR)
  openssl req -config "${CNF_PATH}/ca.cnf" -new -sha256 \
              -subj "/C=OO/O=ACME/OU=ACME Intermediate/CN=${UNIQ_ID_CAI}" \
              -key "ca_i_${UNIQ_ID_CAI}.keys.pem" -out "ca_i_${UNIQ_ID_CAI}.csr.pem"

  # Create Certificate   (valid for ~2 years, after the entire chain of trust expires)
  # CA signs Intermediate
  openssl x509 -req -days 750 -extfile "${CNF_PATH}/ca.cnf" -extensions v3_ca_i \
               -CA $FQ_CA_CERT -CAkey $FQ_CA_KEYS -set_serial ${SERIAL} \
               -in "ca_i_${UNIQ_ID_CAI}.csr.pem" -out "ca_i_${UNIQ_ID_CAI}.crt.pem"

  # Package the Certificate Authority Certificates for distro  (windoze needs this)
  openssl pkcs12 -export -password "pass:password" -inkey "ca_i_${UNIQ_ID_CAI}.keys.pem" \
                 -name "CA Intermediate Mobile Provision" -certfile $FQ_CA_CERT \
                 -in "ca_i_${UNIQ_ID_CAI}.crt.pem" -out "ca_i_${UNIQ_ID_CAI}.p12"

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in "ca_i_${UNIQ_ID_CAI}.crt.pem" > "ca_i_${UNIQ_ID_CAI}.crt.info.txt"

  # create certifiate chain
  cat $FQ_CA_CERT "ca_i_${UNIQ_ID_CAI}.crt.pem" > "ca_cert-chain_${UNIQ_ID_CAI}.crts.pem"
}

#
# Organize the files into logical folders based on serial #
#
ca-i_organize() {
  # organize the client directory
  mkdir -p clients/cfg
  mkdir -p clients/data
  mkdir -p clients/distro
  mkdir -p clients/docs
  mv client*.pem clients/data/
  mv client*.p12 clients/distro/
  mv client*.info.txt clients/docs/
  cp ca_i*.crt.pem clients/cfg/ca_i.crt.pem
  cp ca_i*.keys.pem clients/cfg/ca_i.keys.pem

  # organize the server directory
  mkdir -p servers/cfg
  mkdir -p servers/data
  mkdir -p servers/distro
  mkdir -p servers/docs
  mv server_*.pem servers/data/
  mv server_*.p12 servers/distro/
  mv server_*.info.txt servers/docs/
  cp ca_i*.crt.pem servers/cfg/ca_i.crt.pem
  cp ca_i*.keys.pem servers/cfg/ca_i.keys.pem

  # organize the ca-i directory
  # order matters:  move these files last because they were copied above
  mkdir -p ca-i/data
  mkdir -p ca-i/docs
  mv ca_i*.pem ca-i/data/
  mv ca_i*.info.txt ca-i/docs/
  mv ca_i*.p12 ca-i/
  mv ca_cert-chain*.pem ca-i/
  cp $FQ_CA_DIR/ca_*.crt.pem ca-i/data/
  cp $FQ_CA_DIR/ca_*.info.txt ca-i/docs/
}

#
# Copies all applcations to the Lifecycle package
# 
# Requires: 
#   UNIQ_DIR_LC  : unique string for the Lifecycle directory
#   UNIQ_ID_CAI : unique string for the CA-I
#
ca-i_cp_docs() {
  # CA-I
  cp $CD_ROOT/libs/gen_ca-i.sh $CD_ROOT/$UNIQ_DIR_LC/
  cp $CD_ROOT/libs/pki_funcs.sh $CD_ROOT/$UNIQ_DIR_LC/
  cp $CD_ROOT/docs/README_LC $CD_ROOT/$UNIQ_DIR_LC/README
  cp $CD_ROOT/docs/SERIAL_LC $CD_ROOT/$UNIQ_DIR_LC/cfg/SERIAL
  cp $CD_ROOT/cnf/ca.cnf $CD_ROOT/$UNIQ_DIR_LC/cfg/
  cp $CD_ROOT/$UNIQ_DIR_LC/"ca_${UNIQ_ID_CA}"/ca_*.crt.pem $CD_ROOT/$UNIQ_DIR_LC/cfg/ca.crt.pem
  cp $CD_ROOT/$UNIQ_DIR_LC/"ca_${UNIQ_ID_CA}"/ca_*.keys.pem $CD_ROOT/$UNIQ_DIR_LC/cfg/ca.keys.pem

  # client
  cp $CD_ROOT/libs/gen_client.sh $CD_ROOT/$UNIQ_DIR_LC/distribution/$UNIQ_DIR_CA/clients/
  cp $CD_ROOT/libs/pki_funcs.sh $CD_ROOT/$UNIQ_DIR_LC/distribution/$UNIQ_DIR_CA/clients/
  cp $CD_ROOT/docs/README_C $CD_ROOT/$UNIQ_DIR_LC/distribution/$UNIQ_DIR_CA/clients/README
  cp $CD_ROOT/docs/SERIAL $CD_ROOT/$UNIQ_DIR_LC/distribution/$UNIQ_DIR_CA/clients/cfg/
  cp "${CD_ROOT}/cnf/${ORG_URL}.cnf" $CD_ROOT/$UNIQ_DIR_LC/distribution/$UNIQ_DIR_CA/clients/cfg/

  # server
  cp $CD_ROOT/libs/gen_server.sh $CD_ROOT/$UNIQ_DIR_LC/distribution/$UNIQ_DIR_CA/servers/
  cp $CD_ROOT/libs/pki_funcs.sh $CD_ROOT/$UNIQ_DIR_LC/distribution/$UNIQ_DIR_CA/servers/
  cp $CD_ROOT/docs/README_S $CD_ROOT/$UNIQ_DIR_LC/distribution/$UNIQ_DIR_CA/servers/README
  cp $CD_ROOT/docs/SERIAL $CD_ROOT/$UNIQ_DIR_LC/distribution/$UNIQ_DIR_CA/servers/cfg/
  cp "${CD_ROOT}/cnf/${ORG_URL}.cnf" $CD_ROOT/$UNIQ_DIR_LC/distribution/$UNIQ_DIR_CA/servers/cfg/
}

#
# Generate a Server Certificate
# IN: UNIQ_ID, UNIQ_ID_CA, SERIAL
#
gen_server() {
  ORG_URL=$1 
  UNIQ_ID_CAI=$2
  SERIAL=$3

  UNIQ_ID="${SERIAL}.${ORG_URL}"
  echo_block "Generate Server Certificates  (${UNIQ_ID})"

  openssl genrsa -out "server_${UNIQ_ID}.keys.pem" 4096

  openssl req -new -config $CNF_PATH/${ORG_URL}.cnf -key "server_${UNIQ_ID}.keys.pem" \
              -subj "/C=OO/O=ACME/OU=ACME Standard/CN=${UNIQ_ID}" \
              -out "server_${UNIQ_ID}.csr.pem"

  # CA Intermediate signs Server
  openssl x509 -req -days 365 -extfile $CNF_PATH/${ORG_URL}.cnf -extensions v3_server \
               -CA "ca_i_${UNIQ_ID_CAI}.crt.pem" -CAkey "ca_i_${UNIQ_ID_CAI}.keys.pem" -set_serial ${SERIAL} \
               -in "server_${UNIQ_ID}.csr.pem" -out "server_${UNIQ_ID}.crt.pem"

  # Package the Certificates
  openssl pkcs12 -export -password "pass:password" -inkey "server_${UNIQ_ID}.keys.pem" \
                 -name "Server ${UNIQ_ID} VPN Certificate" -certfile "ca_cert-chain_${UNIQ_ID_CAI}.crts.pem" -caname "server_${UNIQ_ID}@acme.xyz" \
                 -in "server_${UNIQ_ID}.crt.pem" -out "server_${UNIQ_ID}.p12"

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in "server_${UNIQ_ID}.crt.pem" > "server_${UNIQ_ID}.crt.info.txt" 
}

#
# Generate a Client Certificate
# IN: UNIQ_ID, UNIQ_ID_CAI, SERIAL
#
gen_client() {
  ORG_URL=$1 
  UNIQ_ID_CAI=$2
  SERIAL=$3

  UNIQ_ID="${SERIAL}.${ORG_URL}"

  echo_block "Generate Client Certificates  (${UNIQ_ID})"

  openssl genrsa -out "client_${UNIQ_ID}.keys.pem" 4096

  openssl req -new -key "client_${UNIQ_ID}.keys.pem" \
              -subj "/C=OO/O=ACME/OU=ACME Standard/CN=client_${UNIQ_ID}" \
              -out "client_${UNIQ_ID}.csr.pem"
  # CA Intermediate signs Client
  openssl x509 -req -days 365 \
               -CA "ca_i_${UNIQ_ID_CAI}.crt.pem" -CAkey "ca_i_${UNIQ_ID_CAI}.keys.pem" -set_serial ${SERIAL} \
               -in "client_${UNIQ_ID}.csr.pem" -out "client_${UNIQ_ID}.crt.pem"

  # Package the Certificates
  openssl pkcs12 -export -password "pass:password" -inkey "client_${UNIQ_ID}.keys.pem" \
                 -name "Client ${UNIQ_ID} VPN Certificate" -certfile "ca_cert-chain_${UNIQ_ID_CAI}.crts.pem" -caname "client_${UNIQ_ID}@acme.xyz" \
                 -in "client_${UNIQ_ID}.crt.pem" -out "client_${UNIQ_ID}.p12"

  # verify certificate (output to text file for review)
  openssl x509 -noout -text -in "client_${UNIQ_ID}.crt.pem" > "client_${UNIQ_ID}.info.txt"
}

#
# give some info if someone tries to execute this
# echo_block "this script file has only helper functions"

